@@ -5,8 +5,7 @@ use icu_properties::maps::CodePointMapData;
 use icu_provider::AsDeserializingBufferProvider;
 use icu_provider_blob::BlobDataProvider;
 
-use crate::diag::bail;
-use crate::foundations::{Content, NativeElement, SymbolElem, cast, elem, func};
+use crate::foundations::{Content, NativeElement, Str, SymbolElem, cast, elem, func};
 use crate::layout::{Length, Rel};
 use crate::math::Mathy;
 
@@ -87,6 +86,15 @@ impl Accent {
         Self(Self::combine(c).unwrap_or(c))
     }
 
+    /// Tries to select the appropriate combining accent for a string, falling
+    /// back to the string's lone character if there is no corresponding one.
+    ///
+    /// Returns `None` if there isn't one and the string has more than one
+    /// character.
+    pub fn normalize(s: &str) -> Option<Self> {
+        Self::combining(s).or_else(|| s.parse::<char>().ok().map(Self))
+    }
+
     /// Whether this accent is a bottom accent or not.
     pub fn is_bottom(&self) -> bool {
         static COMBINING_CLASS_DATA: LazyLock<CodePointMapData<CanonicalCombiningClass>> =
@@ -110,11 +118,11 @@ impl Accent {
 ///
 /// ```ignore
 /// accents! {
-///     '\u{0300}' | '`' => grave,
-/// //  ^^^^^^^^^    ^^^    ^^^^^
-/// //  |            |      |
-/// //  |            |      +-- The name of the function.
-/// //  |            +--------- The alternative characters that represent the accent.
+///     '\u{0300}', "\u{0300}" | "`" => grave,
+/// //  ^^^^^^^^^   ^^^^^^^^^^^^^^^^    ^^^^^
+/// //  |           |                   |
+/// //  |           |                   +-- The name of the function.
+/// //  |           +--------- The list of strings that normalize to the accent.
 /// //  +---------------------- The primary character that represents the accent.
 /// }
 /// ```
@@ -122,12 +130,18 @@ impl Accent {
 /// When combined with the `Accent::combine` function, accent characters can be normalized
 /// to the primary character.
 macro_rules! accents {
-    ($($primary:literal $(| $alt:literal)* => $name:ident),* $(,)?) => {
+    ($($primary:literal, $($option:literal)|* => $name:ident),* $(,)?) => {
         impl Accent {
             /// Normalize an accent to a combining one.
             pub fn combine(c: char) -> Option<char> {
-                Some(match c {
-                    $($primary $(| $alt)* => $primary,)*
+                Self::combining(c.encode_utf8(&mut [0; 4])).map(|v| v.0)
+            }
+
+            /// Tries to select a well-known combining accent that matches for the
+            /// value.
+            pub fn combining(value: &str) -> Option<Self> {
+                Some(match value {
+                    $($($option)|* => Accent($primary),)*
                     _ => return None,
                 })
             }
@@ -162,33 +176,39 @@ macro_rules! accents {
 
 // Keep it synced with the documenting table above.
 accents! {
-    '\u{0300}' | '`' => grave,
-    '\u{0301}' | '´' => acute,
-    '\u{0302}' | '^' | 'ˆ' => hat,
-    '\u{0303}' | '~' | '∼' | '˜' => tilde,
-    '\u{0304}' | '¯' => macron,
-    '\u{0305}' | '-' | '‾' | '−' => dash,
-    '\u{0306}' | '˘' => breve,
-    '\u{0307}' | '.' | '˙' | '⋅' => dot,
-    '\u{0308}' | '¨' => dot_double,
-    '\u{20db}' => dot_triple,
-    '\u{20dc}' => dot_quad,
-    '\u{030a}' | '∘' | '○' => circle,
-    '\u{030b}' | '˝' => acute_double,
-    '\u{030c}' | 'ˇ' => caron,
-    '\u{20d6}' | '←' => arrow_l,
-    '\u{20d7}' | '→' | '⟶' => arrow,
-    '\u{20e1}' | '↔' | '⟷' => arrow_l_r,
-    '\u{20d0}' | '↼' => harpoon_lt,
-    '\u{20d1}' | '⇀' => harpoon,
+    // Note: Symbols that can have a text presentation must explicitly have that
+    // alternative listed here.
+    '\u{0300}', "\u{0300}" | "`" => grave,
+    '\u{0301}', "\u{0301}" | "´" => acute,
+    '\u{0302}', "\u{0302}" | "^" | "ˆ" => hat,
+    '\u{0303}', "\u{0303}" | "~" | "∼" | "˜" => tilde,
+    '\u{0304}', "\u{0304}" | "¯" => macron,
+    '\u{0305}', "\u{0305}" | "-" | "‾" | "−" => dash,
+    '\u{0306}', "\u{0306}" | "˘" => breve,
+    '\u{0307}', "\u{0307}" | "." | "˙" | "⋅" => dot,
+    '\u{0308}', "\u{0308}" | "¨" => dot_double,
+    '\u{20db}', "\u{20db}" => dot_triple,
+    '\u{20dc}', "\u{20dc}" => dot_quad,
+    '\u{030a}', "\u{030a}" | "∘" | "○" => circle,
+    '\u{030b}', "\u{030b}" | "˝" => acute_double,
+    '\u{030c}', "\u{030c}" | "ˇ" => caron,
+    '\u{20d6}', "\u{20d6}" | "←" => arrow_l,
+    '\u{20d7}', "\u{20d7}" | "→" | "⟶" => arrow,
+    '\u{20e1}', "\u{20e1}" | "↔" | "↔\u{fe0e}" | "⟷" => arrow_l_r,
+    '\u{20d0}', "\u{20d0}" | "↼" => harpoon_lt,
+    '\u{20d1}', "\u{20d1}" | "⇀" => harpoon,
 }
 
 cast! {
     Accent,
     self => self.0.into_value(),
-    v: char => Self::new(v),
-    v: Content => match v.to_packed::<SymbolElem>().and_then(|elem| elem.text.parse::<char>().ok()) {
-        Some(c) => Self::new(c),
-        _ => bail!("expected a single-codepoint symbol"),
-    },
+    // The string cast handles
+    // - strings: `accent(a, "↔")`
+    // - symbol values: `accent(a, <->)`
+    // - shorthands: `accent(a, arrow.l.r)`
+    v: Str => Self::normalize(&v).ok_or("expected exactly one character")?,
+    // The content cast is for accent uses like `accent(a, ↔)`
+    v: Content => v.to_packed::<SymbolElem>()
+        .and_then(|elem| Accent::normalize(&elem.text))
+        .ok_or("expected a single-codepoint symbol")?,
 }
