@@ -1,16 +1,15 @@
 use ecow::EcoString;
-use ttf_parser::OutlineBuilder;
-use typst_library::layout::{Abs, Ratio, Size, Transform};
+use typst_library::layout::{Abs, Point, Ratio, Size, Transform};
 use typst_library::visualize::{
     Curve, CurveItem, FixedStroke, Geometry, LineCap, LineJoin, Paint, RelativeTo, Shape,
 };
 
 use crate::paint::ColorEncode;
-use crate::{SVGRenderer, State, SvgPathBuilder};
+use crate::{SVGRenderer, State, SvgMatrix, SvgPathBuilder};
 
-impl SVGRenderer {
+impl SVGRenderer<'_> {
     /// Render a shape element.
-    pub(super) fn render_shape(&mut self, state: State, shape: &Shape) {
+    pub(super) fn render_shape(&mut self, state: &State, shape: &Shape) {
         self.xml.start_element("path");
         self.xml.write_attribute("class", "typst-shape");
 
@@ -33,6 +32,10 @@ impl SVGRenderer {
             );
         }
 
+        if !state.transform.is_identity() {
+            self.xml.write_attribute("transform", &SvgMatrix(state.transform));
+        }
+
         let path = convert_geometry_to_path(&shape.geometry);
         self.xml.write_attribute("d", &path);
         self.xml.end_element();
@@ -41,7 +44,7 @@ impl SVGRenderer {
     /// Calculate the transform of the shape's fill or stroke.
     fn shape_paint_transform(
         &self,
-        state: State,
+        state: &State,
         paint: &Paint,
         shape: &Shape,
     ) -> Transform {
@@ -78,7 +81,7 @@ impl SVGRenderer {
     }
 
     /// Calculate the size of the shape's fill.
-    fn shape_fill_size(&self, state: State, paint: &Paint, shape: &Shape) -> Size {
+    fn shape_fill_size(&self, state: &State, paint: &Paint, shape: &Shape) -> Size {
         let mut shape_size = shape.geometry.bbox_size();
         // Edge cases for strokes.
         if shape_size.x.to_pt() == 0.0 {
@@ -155,7 +158,9 @@ impl SVGRenderer {
 /// Convert a geometry to an SVG path.
 #[comemo::memoize]
 fn convert_geometry_to_path(geometry: &Geometry) -> EcoString {
-    let mut builder = SvgPathBuilder::default();
+    let mut builder =
+        SvgPathBuilder::with_translate(Point::new(Abs::zero(), Abs::zero()));
+
     match geometry {
         Geometry::Line(t) => {
             builder.move_to(0.0, 0.0);
@@ -166,13 +171,15 @@ fn convert_geometry_to_path(geometry: &Geometry) -> EcoString {
             let y = rect.y.to_pt() as f32;
             builder.rect(x, y);
         }
-        Geometry::Curve(p) => return convert_curve(p),
+        Geometry::Curve(p) => {
+            return convert_curve(Point::new(Abs::zero(), Abs::zero()), p);
+        }
     };
-    builder.0
+    builder.path
 }
 
-pub fn convert_curve(curve: &Curve) -> EcoString {
-    let mut builder = SvgPathBuilder::default();
+pub fn convert_curve(initial_point: Point, curve: &Curve) -> EcoString {
+    let mut builder = SvgPathBuilder::with_translate(initial_point);
     for item in &curve.0 {
         match item {
             CurveItem::Move(m) => builder.move_to(m.x.to_pt() as f32, m.y.to_pt() as f32),
@@ -188,5 +195,5 @@ pub fn convert_curve(curve: &Curve) -> EcoString {
             CurveItem::Close => builder.close(),
         }
     }
-    builder.0
+    builder.path
 }
