@@ -3,9 +3,8 @@ use std::ops::{Deref, Range};
 use std::rc::Rc;
 use std::sync::Arc;
 
-use ecow::{eco_format, eco_vec, EcoString, EcoVec};
+use ecow::{EcoString, EcoVec, eco_format, eco_vec};
 
-use crate::ast::AstNode;
 use crate::{FileId, Span, SyntaxKind};
 
 /// A node in the untyped syntax tree.
@@ -119,26 +118,6 @@ impl SyntaxNode {
         }
     }
 
-    /// Whether the node can be cast to the given AST node.
-    pub fn is<'a, T: AstNode<'a>>(&'a self) -> bool {
-        self.cast::<T>().is_some()
-    }
-
-    /// Try to convert the node to a typed AST node.
-    pub fn cast<'a, T: AstNode<'a>>(&'a self) -> Option<T> {
-        T::from_untyped(self)
-    }
-
-    /// Cast the first child that can cast to the AST type `T`.
-    pub fn cast_first_match<'a, T: AstNode<'a>>(&'a self) -> Option<T> {
-        self.children().find_map(Self::cast)
-    }
-
-    /// Cast the last child that can cast to the AST type `T`.
-    pub fn cast_last_match<'a, T: AstNode<'a>>(&'a self) -> Option<T> {
-        self.children().rev().find_map(Self::cast)
-    }
-
     /// Whether the node or its children contain an error.
     pub fn erroneous(&self) -> bool {
         match &self.0 {
@@ -753,11 +732,10 @@ impl<'a> LinkedNode<'a> {
                 // sibling's span number is larger than the target span's number.
                 if children
                     .peek()
-                    .map_or(true, |next| next.span().number() > span.number())
+                    .is_none_or(|next| next.span().number() > span.number())
+                    && let Some(found) = child.find(span)
                 {
-                    if let Some(found) = child.find(span) {
-                        return Some(found);
-                    }
+                    return Some(found);
                 }
             }
         }
@@ -780,11 +758,7 @@ impl LinkedNode<'_> {
         let node = parent.node.children().nth(index)?;
         let offset = self.offset - node.len();
         let prev = Self { node, parent: self.parent.clone(), index, offset };
-        if prev.kind().is_trivia() {
-            prev.prev_sibling()
-        } else {
-            Some(prev)
-        }
+        if prev.kind().is_trivia() { prev.prev_sibling() } else { Some(prev) }
     }
 
     /// Get the next non-trivia sibling node.
@@ -794,11 +768,7 @@ impl LinkedNode<'_> {
         let node = parent.node.children().nth(index)?;
         let offset = self.offset + self.node.len();
         let next = Self { node, parent: self.parent.clone(), index, offset };
-        if next.kind().is_trivia() {
-            next.next_sibling()
-        } else {
-            Some(next)
-        }
+        if next.kind().is_trivia() { next.next_sibling() } else { Some(next) }
     }
 
     /// Get the kind of this node's parent.
