@@ -1,13 +1,16 @@
-use std::collections::{HashMap, HashSet};
 use std::fmt::{self, Display, Formatter};
 use std::ops::Range;
 use std::path::{Path, PathBuf};
 use std::str::FromStr;
+use std::sync::LazyLock;
 
-use ecow::{eco_format, EcoString};
-use once_cell::sync::Lazy;
-use typst::syntax::package::PackageVersion;
-use typst::syntax::{is_id_continue, is_ident, is_newline, FileId, Source, VirtualPath};
+use bitflags::bitflags;
+use ecow::{EcoString, eco_format};
+use rustc_hash::{FxHashMap, FxHashSet};
+use typst_syntax::package::PackageVersion;
+use typst_syntax::{
+    FileId, Lines, Source, VirtualPath, is_id_continue, is_ident, is_newline,
+};
 use unscanny::Scanner;
 
 /// Collects all tests from all files.
@@ -23,14 +26,15 @@ pub fn collect() -> Result<(Vec<Test>, usize), Vec<TestParseError>> {
 pub struct Test {
     pub pos: FilePos,
     pub name: EcoString,
+    pub attrs: Attrs,
     pub source: Source,
     pub notes: Vec<Note>,
-    pub large: bool,
 }
 
 impl Display for Test {
     fn fmt(&self, f: &mut Formatter) -> fmt::Result {
-        write!(f, "{} ({})", self.name, self.pos)
+        // underline path
+        write!(f, "{} (\x1B[4m{}\x1B[0m)", self.name, self.pos)
     }
 }
 
@@ -57,6 +61,62 @@ impl Display for FilePos {
     }
 }
 
+bitflags! {
+    #[derive(Copy, Clone)]
+    struct AttrFlags: u8 {
+        const RENDER = 1 << 0;
+        const HTML = 1 << 1;
+        const PDFTAGS = 1 << 2;
+        const LARGE = 1 << 3;
+        const NOPDFUA = 1 << 4;
+    }
+}
+
+impl AttrFlags {
+    const NON_RENDER: Self = Self::HTML.union(Self::PDFTAGS);
+
+    pub fn targets(self) -> Targets {
+        let mut targets = Targets::empty();
+        if self.contains(Self::RENDER) || (self & Self::NON_RENDER).is_empty() {
+            targets |= Targets::RENDER;
+        }
+        if self.contains(Self::HTML) {
+            targets |= Targets::HTML;
+        }
+        if self.contains(Self::PDFTAGS) {
+            targets |= Targets::PDFTAGS;
+        }
+        targets
+    }
+}
+
+#[derive(Debug, Default, Copy, Clone, Eq, PartialEq)]
+pub struct Attrs {
+    pub large: bool,
+    pub pdf_ua: bool,
+    pub targets: Targets,
+}
+
+bitflags! {
+    #[derive(Debug, Default, Copy, Clone, Eq, PartialEq)]
+    pub struct Targets: u8 {
+        const RENDER = 0x1;
+        const HTML = 0x2;
+        const PDFTAGS = 0x4;
+    }
+}
+
+impl Targets {
+    pub fn from_file_extension(ext: &str) -> Option<Self> {
+        Some(match ext {
+            "png" => Self::RENDER,
+            "html" => Self::HTML,
+            "yml" => Self::PDFTAGS,
+            _ => return None,
+        })
+    }
+}
+
 /// The size of a file.
 pub struct FileSize(pub usize);
 
@@ -70,6 +130,8 @@ impl Display for FileSize {
 pub struct Note {
     pub pos: FilePos,
     pub kind: NoteKind,
+    /// The file [`Self::range`] belongs to.
+    pub file: FileId,
     pub range: Option<Range<usize>>,
     pub message: String,
 }
@@ -109,8 +171,7 @@ impl Display for NoteKind {
 struct Collector {
     tests: Vec<Test>,
     errors: Vec<TestParseError>,
-    seen: HashMap<EcoString, FilePos>,
-    large: HashSet<EcoString>,
+    seen: FxHashMap<EcoString, (FilePos, Attrs)>,
     skipped: usize,
 }
 
@@ -120,8 +181,7 @@ impl Collector {
         Self {
             tests: vec![],
             errors: vec![],
-            seen: HashMap::new(),
-            large: HashSet::new(),
+            seen: FxHashMap::default(),
             skipped: 0,
         }
     }
@@ -143,7 +203,7 @@ impl Collector {
         for entry in walkdir::WalkDir::new(crate::SUITE_PATH).sort_by_file_name() {
             let entry = entry.unwrap();
             let path = entry.path();
-            if !path.extension().is_some_and(|ext| ext == "typ") {
+            if path.extension().is_none_or(|ext| ext != "typ") {
                 continue;
             }
 
@@ -156,33 +216,43 @@ impl Collector {
         }
     }
 
-    /// Walks through all reference images and ensure that a test exists for
+    /// Walks through all reference output and ensures that a test exists for
     /// each one.
     fn walk_references(&mut self) {
         for entry in walkdir::WalkDir::new(crate::REF_PATH).sort_by_file_name() {
             let entry = entry.unwrap();
             let path = entry.path();
-            if !path.extension().is_some_and(|ext| ext == "png") {
+            let Some(file_target) = path.extension().and_then(|ext| {
+                let str = ext.to_str()?;
+                Targets::from_file_extension(str)
+            }) else {
                 continue;
-            }
+            };
 
             let stem = path.file_stem().unwrap().to_string_lossy();
             let name = &*stem;
 
-            let Some(pos) = self.seen.get(name) else {
+            let Some((pos, attrs)) = self.seen.get(name) else {
                 self.errors.push(TestParseError {
                     pos: FilePos::new(path, 0),
-                    message: "dangling reference image".into(),
+                    message: "dangling reference output".into(),
                 });
                 continue;
             };
 
+            if !attrs.targets.contains(file_target) {
+                self.errors.push(TestParseError {
+                    pos: FilePos::new(path, 0),
+                    message: "dangling reference output".into(),
+                });
+            }
+
             let len = path.metadata().unwrap().len() as usize;
-            if !self.large.contains(name) && len > crate::REF_LIMIT {
+            if !attrs.large && len > crate::REF_LIMIT {
                 self.errors.push(TestParseError {
                     pos: pos.clone(),
                     message: format!(
-                        "reference image size exceeds {}, but the test is not marked as `// LARGE`",
+                        "reference output size exceeds {}, but the test is not marked as `large`",
                         FileSize(crate::REF_LIMIT),
                     ),
                 });
@@ -218,6 +288,7 @@ impl<'a> Parser<'a> {
 
         while !self.s.done() {
             let mut name = EcoString::new();
+            let mut attrs = Attrs::default();
             let mut notes = vec![];
             if self.s.eat_if("---") {
                 self.s.eat_while(' ');
@@ -228,8 +299,8 @@ impl<'a> Parser<'a> {
                     self.error("expected test name");
                 } else if !is_ident(&name) {
                     self.error(format!("test name `{name}` is not a valid identifier"));
-                } else if !self.s.eat_if("---") {
-                    self.error("expected closing ---");
+                } else {
+                    attrs = self.parse_attrs();
                 }
             } else {
                 self.error("expected opening ---");
@@ -247,7 +318,7 @@ impl<'a> Parser<'a> {
             self.test_start_line = self.line;
 
             let pos = FilePos::new(self.path, self.test_start_line);
-            self.collector.seen.insert(name.clone(), pos.clone());
+            self.collector.seen.insert(name.clone(), (pos.clone(), attrs));
 
             while !self.s.done() && !self.s.at("---") {
                 self.s.eat_until(is_newline);
@@ -257,10 +328,6 @@ impl<'a> Parser<'a> {
             }
 
             let text = self.s.from(start);
-            let large = text.starts_with("// LARGE");
-            if large {
-                self.collector.large.insert(name.clone());
-            }
 
             if !selected(&name, self.path.canonicalize().unwrap()) {
                 self.collector.skipped += 1;
@@ -285,7 +352,38 @@ impl<'a> Parser<'a> {
                 }
             }
 
-            self.collector.tests.push(Test { pos, name, source, notes, large });
+            self.collector.tests.push(Test { pos, name, source, notes, attrs });
+        }
+    }
+
+    fn parse_attrs(&mut self) -> Attrs {
+        let mut parsed = AttrFlags::empty();
+        while !self.s.eat_if("---") {
+            let attr = self.s.eat_until(char::is_whitespace);
+            let flag = match attr {
+                "large" => AttrFlags::LARGE,
+                "html" => AttrFlags::HTML,
+                "render" => AttrFlags::RENDER,
+                "pdftags" => AttrFlags::PDFTAGS,
+                "nopdfua" => AttrFlags::NOPDFUA,
+                found => {
+                    self.error(format!(
+                        "expected attribute or closing ---, found `{found}`"
+                    ));
+                    break;
+                }
+            };
+            if parsed.contains(flag) {
+                self.error(format!("duplicate attribute `{attr}`"));
+            }
+            parsed.insert(flag);
+            self.s.eat_while(' ');
+        }
+
+        Attrs {
+            large: parsed.contains(AttrFlags::LARGE),
+            pdf_ua: !parsed.contains(AttrFlags::NOPDFUA),
+            targets: parsed.targets(),
         }
     }
 
@@ -314,9 +412,28 @@ impl<'a> Parser<'a> {
         let kind: NoteKind = head.parse().ok()?;
         self.s.eat_if(' ');
 
+        let mut file = None;
+        if self.s.eat_if('"') {
+            let path = self.s.eat_until(|c| is_newline(c) || c == '"');
+            if !self.s.eat_if('"') {
+                self.error("expected closing quote after file path");
+                return None;
+            }
+
+            let vpath = VirtualPath::new(path);
+            file = Some(FileId::new(None, vpath));
+
+            self.s.eat_if(' ');
+        }
+
         let mut range = None;
         if self.s.at('-') || self.s.at(char::is_numeric) {
-            range = self.parse_range(source);
+            if let Some(file) = file {
+                range = self.parse_range_external(file);
+            } else {
+                range = self.parse_range(source);
+            }
+
             if range.is_none() {
                 self.error("range is malformed");
                 return None;
@@ -332,11 +449,78 @@ impl<'a> Parser<'a> {
         Some(Note {
             pos: FilePos::new(self.path, self.line),
             kind,
+            file: file.unwrap_or(source.id()),
             range,
             message,
         })
     }
 
+    #[cfg(not(feature = "default"))]
+    fn parse_range_external(&mut self, _file: FileId) -> Option<Range<usize>> {
+        panic!("external file ranges are not expected when testing `typst_syntax`");
+    }
+
+    /// Parse a range in an external file, optionally abbreviated as just a position
+    /// if the range is empty.
+    #[cfg(feature = "default")]
+    fn parse_range_external(&mut self, file: FileId) -> Option<Range<usize>> {
+        use typst::foundations::Bytes;
+
+        use crate::world::{read, system_path};
+
+        let path = match system_path(file) {
+            Ok(path) => path,
+            Err(err) => {
+                self.error(err.to_string());
+                return None;
+            }
+        };
+
+        let bytes = match read(&path) {
+            Ok(data) => Bytes::new(data),
+            Err(err) => {
+                self.error(err.to_string());
+                return None;
+            }
+        };
+
+        let start = self.parse_line_col()?;
+        let lines = Lines::try_from(&bytes).expect(
+            "errors shouldn't be annotated for files \
+            that aren't human readable (not valid utf-8)",
+        );
+        let range = if self.s.eat_if('-') {
+            let (line, col) = start;
+            let start = lines.line_column_to_byte(line, col);
+            let (line, col) = self.parse_line_col()?;
+            let end = lines.line_column_to_byte(line, col);
+            Option::zip(start, end).map(|(a, b)| a..b)
+        } else {
+            let (line, col) = start;
+            lines.line_column_to_byte(line, col).map(|i| i..i)
+        };
+        if range.is_none() {
+            self.error("range is out of bounds");
+        }
+        range
+    }
+
+    /// Parses absolute `line:column` indices in an external file.
+    fn parse_line_col(&mut self) -> Option<(usize, usize)> {
+        let line = self.parse_number()?;
+        if !self.s.eat_if(':') {
+            self.error("positions in external files always require both `<line>:<col>`");
+            return None;
+        }
+        let col = self.parse_number()?;
+        if line < 0 || col < 0 {
+            self.error("line and column numbers must be positive");
+            return None;
+        }
+
+        Some(((line as usize).saturating_sub(1), (col as usize).saturating_sub(1)))
+    }
+
     /// Parse a range, optionally abbreviated as just a position if the range
     /// is empty.
     fn parse_range(&mut self, source: &Source) -> Option<Range<usize>> {
@@ -362,13 +546,13 @@ impl<'a> Parser<'a> {
         let line_idx = (line_idx_in_test + comments).checked_add_signed(line_delta)?;
         let column_idx = if column < 0 {
             // Negative column index is from the back.
-            let range = source.line_to_range(line_idx)?;
+            let range = source.lines().line_to_range(line_idx)?;
             text[range].chars().count().saturating_add_signed(column)
         } else {
             usize::try_from(column).ok()?.checked_sub(1)?
         };
 
-        source.line_column_to_byte(line_idx, column_idx)
+        source.lines().line_column_to_byte(line_idx, column_idx)
     }
 
     /// Parse a number.
@@ -390,7 +574,7 @@ impl<'a> Parser<'a> {
 
 /// Whether a test is within the selected set to run.
 fn selected(name: &str, abs: PathBuf) -> bool {
-    static SKIPPED: Lazy<HashSet<&'static str>> = Lazy::new(|| {
+    static SKIPPED: LazyLock<FxHashSet<&'static str>> = LazyLock::new(|| {
         String::leak(std::fs::read_to_string(crate::SKIP_PATH).unwrap())
             .lines()
             .map(|line| line.trim())
@@ -411,11 +595,7 @@ fn selected(name: &str, abs: PathBuf) -> bool {
     let patterns = &crate::ARGS.pattern;
     patterns.is_empty()
         || patterns.iter().any(|pattern: &regex::Regex| {
-            if exact {
-                name == pattern.as_str()
-            } else {
-                pattern.is_match(name)
-            }
+            if exact { name == pattern.as_str() } else { pattern.is_match(name) }
         })
 }
 
